@@ -52,7 +52,10 @@
 #include "BotanGOST.h"
 #include "BotanGOSTR3411.h"
 #endif
-#include "BotanHMAC.h"
+#include "BotanMAC.h"
+#ifdef WITH_EDDSA
+#include "BotanEDDSA.h"
+#endif
 
 #include <botan/init.h>
 
@@ -175,6 +178,10 @@ AsymmetricAlgorithm* BotanCryptoFactory::getAsymmetricAlgorithm(AsymAlgo::Type a
 #ifdef WITH_GOST
 		case AsymAlgo::GOST:
 			return new BotanGOST();
+#endif
+#ifdef WITH_EDDSA
+		case AsymAlgo::EDDSA:
+			return new BotanEDDSA();
 #endif
 		default:
 			// No algorithm implementation is available
@@ -240,6 +247,10 @@ MacAlgorithm* BotanCryptoFactory::getMacAlgorithm(MacAlgo::Type algorithm)
 		case MacAlgo::HMAC_GOST:
 			return new BotanHMACGOSTR3411();
 #endif
+		case MacAlgo::CMAC_DES:
+			return new BotanCMACDES();
+		case MacAlgo::CMAC_AES:
+			return new BotanCMACAES();
 		default:
 			// No algorithm implementation is available
 			ERROR_MSG("Unknown algorithm '%i'", algorithm);
