@@ -40,6 +40,7 @@
 #include "BotanRSAKeyPair.h"
 #include <algorithm>
 #include <botan/rsa.h>
+#include <botan/version.h>
 #include <sstream>
 
 // Constructor
@@ -71,6 +72,15 @@ bool BotanRSA::sign(PrivateKey* privateKey, const ByteString& dataToSign,
 		case AsymMech::RSA_PKCS:
 			emsa = "EMSA3(Raw)";
 			break;
+#ifdef WITH_RAW_PSS
+		case AsymMech::RSA_PKCS_PSS:
+			emsa = getCipherRawPss(privateKey->getBitLength(), dataToSign.size(), param, paramLen);
+			if (emsa == "")
+			{
+				return false;
+			}
+			break;
+#endif
 		default:
 			// Call default implementation
 			return AsymmetricAlgorithm::sign(privateKey, dataToSign, signature, mechanism, param, paramLen);
@@ -96,7 +106,12 @@ bool BotanRSA::sign(PrivateKey* privateKey, const ByteString& dataToSign,
 
 	try
 	{
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,33)
+		BotanRNG* rng = (BotanRNG*)BotanCryptoFactory::i()->getRNG();
+		signer = new Botan::PK_Signer(*botanKey, *rng->getRNG(), emsa);
+#else
 		signer = new Botan::PK_Signer(*botanKey, emsa);
+#endif
 		// Should we add DISABLE_FAULT_PROTECTION? Makes this operation faster.
 	}
 	catch (...)
@@ -107,7 +122,7 @@ bool BotanRSA::sign(PrivateKey* privateKey, const ByteString& dataToSign,
 	}
 
 	// Perform the signature operation
-#if BOTAN_VERSION_MINOR == 11
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,0)
 	std::vector<Botan::byte> signResult;
 #else
 	Botan::SecureVector<Botan::byte> signResult;
@@ -129,7 +144,7 @@ bool BotanRSA::sign(PrivateKey* privateKey, const ByteString& dataToSign,
 
 	// Return the result
 	signature.resize(signResult.size());
-#if BOTAN_VERSION_MINOR == 11
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,0)
 	memcpy(&signature[0], signResult.data(), signResult.size());
 #else
 	memcpy(&signature[0], signResult.begin(), signResult.size());
@@ -321,7 +336,12 @@ bool BotanRSA::signInit(PrivateKey* privateKey, const AsymMech::Type mechanism,
 
 	try
 	{
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,33)
+		BotanRNG* rng = (BotanRNG*)BotanCryptoFactory::i()->getRNG();
+		signer = new Botan::PK_Signer(*botanKey, *rng->getRNG(), emsa);
+#else
 		signer = new Botan::PK_Signer(*botanKey, emsa);
+#endif
 		// Should we add DISABLE_FAULT_PROTECTION? Makes this operation faster.
 	}
 	catch (...)
@@ -376,7 +396,7 @@ bool BotanRSA::signFinal(ByteString& signature)
 	}
 
 	// Perform the signature operation
-#if BOTAN_VERSION_MINOR == 11
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,0)
 	std::vector<Botan::byte> signResult;
 #else
 	Botan::SecureVector<Botan::byte> signResult;
@@ -398,7 +418,7 @@ bool BotanRSA::signFinal(ByteString& signature)
 
 	// Return the result
 	signature.resize(signResult.size());
-#if BOTAN_VERSION_MINOR == 11
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,0)
 	memcpy(&signature[0], signResult.data(), signResult.size());
 #else
 	memcpy(&signature[0], signResult.begin(), signResult.size());
@@ -425,6 +445,15 @@ bool BotanRSA::verify(PublicKey* publicKey, const ByteString& originalData,
 		case AsymMech::RSA_PKCS:
 			emsa = "EMSA3(Raw)";
 			break;
+#ifdef WITH_RAW_PSS
+		case AsymMech::RSA_PKCS_PSS:
+			emsa = getCipherRawPss(publicKey->getBitLength(), originalData.size(), param, paramLen);
+			if (emsa == "")
+			{
+				return false;
+			}
+			break;
+#endif
 		default:
 			// Call the generic function
 			return AsymmetricAlgorithm::verify(publicKey, originalData, signature, mechanism, param, paramLen);
@@ -783,7 +812,12 @@ bool BotanRSA::encrypt(PublicKey* publicKey, const ByteString& data,
 	Botan::PK_Encryptor_EME* encryptor = NULL;
 	try
 	{
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,33)
+		BotanRNG* rng = (BotanRNG*)BotanCryptoFactory::i()->getRNG();
+		encryptor = new Botan::PK_Encryptor_EME(*botanKey, *rng->getRNG(), eme);
+#else
 		encryptor = new Botan::PK_Encryptor_EME(*botanKey, eme);
+#endif
 	}
 	catch (...)
 	{
@@ -793,7 +827,7 @@ bool BotanRSA::encrypt(PublicKey* publicKey, const ByteString& data,
 	}
 
 	// Perform the encryption operation
-#if BOTAN_VERSION_MINOR == 11
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,0)
 	std::vector<Botan::byte> encResult;
 #else
 	Botan::SecureVector<Botan::byte> encResult;
@@ -814,7 +848,7 @@ bool BotanRSA::encrypt(PublicKey* publicKey, const ByteString& data,
 
 	// Return the result
 	encryptedData.resize(encResult.size());
-#if BOTAN_VERSION_MINOR == 11
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,0)
 	memcpy(&encryptedData[0], encResult.data(), encResult.size());
 #else
 	memcpy(&encryptedData[0], encResult.begin(), encResult.size());
@@ -869,7 +903,12 @@ bool BotanRSA::decrypt(PrivateKey* privateKey, const ByteString& encryptedData,
 	Botan::PK_Decryptor_EME* decryptor = NULL;
 	try
 	{
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,33)
+		BotanRNG* rng = (BotanRNG*)BotanCryptoFactory::i()->getRNG();
+		decryptor = new Botan::PK_Decryptor_EME(*botanKey, *rng->getRNG(), eme);
+#else
 		decryptor = new Botan::PK_Decryptor_EME(*botanKey, eme);
+#endif
 	}
 	catch (...)
 	{
@@ -879,7 +918,7 @@ bool BotanRSA::decrypt(PrivateKey* privateKey, const ByteString& encryptedData,
 	}
 
 	// Perform the decryption operation
-#if BOTAN_VERSION_MINOR == 11
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,0)
 	Botan::secure_vector<Botan::byte> decResult;
 #else
 	Botan::SecureVector<Botan::byte> decResult;
@@ -904,7 +943,7 @@ bool BotanRSA::decrypt(PrivateKey* privateKey, const ByteString& encryptedData,
 		int modSize = pk->getN().size();
 		int decSize = decResult.size();
 		data.resize(modSize);
-#if BOTAN_VERSION_MINOR == 11
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,0)
 		memcpy(&data[0] + modSize - decSize, decResult.data(), decSize);
 #else
 		memcpy(&data[0] + modSize - decSize, decResult.begin(), decSize);
@@ -913,7 +952,7 @@ bool BotanRSA::decrypt(PrivateKey* privateKey, const ByteString& encryptedData,
 	else
 	{
 		data.resize(decResult.size());
-#if BOTAN_VERSION_MINOR == 11
+#if BOTAN_VERSION_CODE >= BOTAN_VERSION_CODE_FOR(1,11,0)
 		memcpy(&data[0], decResult.data(), decResult.size());
 #else
 		memcpy(&data[0], decResult.begin(), decResult.size());
@@ -1121,3 +1160,60 @@ bool BotanRSA::reconstructParameters(AsymmetricParameters** ppParams, ByteString
 	return true;
 }
 
+#ifdef WITH_RAW_PSS
+std::string BotanRSA::getCipherRawPss(size_t bitLength, size_t dataSize, const void* param, const size_t paramLen)
+{
+	if (param == NULL || paramLen != sizeof(RSA_PKCS_PSS_PARAMS))
+	{
+		ERROR_MSG("Invalid parameters");
+		return "";
+	}
+
+	std::string hashStr = "";
+	size_t allowedLen = 0;
+	switch (((RSA_PKCS_PSS_PARAMS*) param)->hashAlg)
+	{
+		case HashAlgo::SHA1:
+			hashStr = "SHA-160";
+			allowedLen = 20;
+			break;
+		case HashAlgo::SHA224:
+			hashStr = "SHA-224";
+			allowedLen = 28;
+			break;
+		case HashAlgo::SHA256:
+			hashStr = "SHA-256";
+			allowedLen = 32;
+			break;
+		case HashAlgo::SHA384:
+			hashStr = "SHA-384";
+			allowedLen = 48;
+			break;
+		case HashAlgo::SHA512:
+			hashStr = "SHA-512";
+			allowedLen = 64;
+			break;
+		default:
+			ERROR_MSG("Invalid hash parameter");
+			return "";
+	}
+
+	if (dataSize != allowedLen)
+	{
+		ERROR_MSG("Data to sign does not match expected (%d) for RSA PSS", (int)allowedLen);
+		return "";
+	}
+
+	size_t sLen = ((RSA_PKCS_PSS_PARAMS*) param)->sLen;
+	if (sLen > ((bitLength+6)/8-2-20))
+	{
+		ERROR_MSG("sLen (%lu) is too large for current key size (%lu)",
+			  (unsigned long)sLen, bitLength);
+		return "";
+	}
+
+	std::ostringstream request;
+	request << "PSSR_Raw(" << hashStr << ",MGF1," << sLen << ")";
+	return request.str();
+}
+#endif
